package ac.ed.lurg.country.gams;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Vector;

import com.gams.api.GAMSDatabase;
import com.gams.api.GAMSGlobals;
import com.gams.api.GAMSJob;
import com.gams.api.GAMSOptions;
import com.gams.api.GAMSParameter;
import com.gams.api.GAMSParameterRecord;
import com.gams.api.GAMSVariable;
import com.gams.api.GAMSVariableRecord;
import com.gams.api.GAMSWorkspace;
import com.gams.api.GAMSWorkspaceInfo;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.types.CommodityType;
import ac.ed.lurg.utils.LogWriter;

public class GamsDemandOptimiser {

	private GamsDemandInput inputData;

	public GamsDemandOptimiser(GamsDemandInput inputData) {
		this.inputData = inputData;
	}
	
	public GamsDemandOutput getDemandPc() {
		File workingDirectory = new File(ModelConfig.TEMP_DIR);
		workingDirectory.mkdir();
		GAMSWorkspaceInfo  wsInfo  = new GAMSWorkspaceInfo();
		wsInfo.setWorkingDirectory(workingDirectory.getAbsolutePath());

		GAMSWorkspace ws = new GAMSWorkspace(wsInfo);
		GAMSDatabase inDB = ws.addDatabase();
		setupInDB(inDB);

		GAMSJob gamsJob = ws.addJobFromFile(ModelConfig.DEMAND_GAMS_MODEL);
		GAMSOptions opt = ws.addOptions();

		opt.defines("gdx_parameters", new File(ModelConfig.DEMAND_PARAM_FILE).getAbsolutePath());
		opt.defines("gdx_prices_and_gdp", inDB.getName());

		long startTime = System.currentTimeMillis();
		gamsJob.run(opt, inDB);
		
		LogWriter.println("Took " + (System.currentTimeMillis() - startTime) + " ms to run");
		return handleResults(gamsJob.OutDB());
	}

	private void setupInDB(GAMSDatabase inDB) {
		GAMSParameter gdpPcP = inDB.addParameter("gdp_pc", 0);
		double gdpPc = inputData.getGdpPc();
		LogWriter.println(String.format("gdp_pc = %.5f", gdpPc)); 
		setGamsParamValue(gdpPcP.addRecord(), gdpPc, 5);
		GamsDemandOutput previousGamsDemands = inputData.getPreviousDemands();
		
		GAMSParameter prevUP = inDB.addParameter("previousU", 0);
		double previousUtility = previousGamsDemands == null ? 0 : previousGamsDemands.getUtility();//-16.3829624069452 + 4.3659621833299 * Math.log10(gdpPc);
		LogWriter.println(String.format("previousU = %.5f, estimated = %.5f", previousUtility, -16.3829624069452 + 4.3659621833299 * Math.log10(gdpPc))); 
		setGamsParamValue(prevUP.addRecord(), previousUtility, 5);
		
		GAMSParameter maxIncomePropFoodSpendP = inDB.addParameter("maxIncomePropFoodSpend", 0);
		setGamsParamValue(maxIncomePropFoodSpendP.addRecord(), ModelConfig.MAX_INCOME_PROPORTION_FOOD_SPEND, 4);

		LogWriter.println("\nPrice");	
		GAMSParameter priceP = inDB.addParameter("price", 1);	
		GAMSParameter prevSubsP = inDB.addParameter("previousSubs", 1);
		GAMSParameter prevDiscP = inDB.addParameter("previousDisc", 1);

		for (Map.Entry<CommodityType, Double> entry : inputData.getPrices().entrySet()) {
			CommodityType comm = entry.getKey();
			double priceComm = entry.getValue();
			doCommodity(inDB, priceP, prevSubsP, prevDiscP, gdpPc, previousGamsDemands, comm, priceComm, 
					comm==CommodityType.NONFOOD ? 5 : 0.05, comm==CommodityType.NONFOOD ? 0.5 : 0.05);
		}
	}

	private void doCommodity(GAMSDatabase inDB, GAMSParameter priceP, GAMSParameter prevSubsP, GAMSParameter prevDiscP, double gdpPc, 
			GamsDemandOutput previousGamsDemands, CommodityType comm, double price, double defaultSubs, double defaultDisc) {
		
		Vector<String> v = new Vector<String>();
		v.add(comm.getGamsName());
		double prevSubs = defaultSubs, prevDisc = defaultDisc;
		
		if (previousGamsDemands != null) {
			GamsCommodityDemand demand = previousGamsDemands.getGamsDemands(comm);
			prevSubs = demand.getSubsistence();
			prevDisc = demand.getDiscretionary();
		}
		
		LogWriter.println(String.format("%14s:  price=%.4f, previousSubs=%.4f, previousDisc=%.4f", comm.getGamsName(), price, prevSubs, prevDisc));
		setGamsParamValue(priceP.addRecord(v), price, 4);
		setGamsParamValue(prevSubsP.addRecord(v), prevSubs, 4);
		setGamsParamValue(prevDiscP.addRecord(v), prevDisc, 4);
	}

	private double setGamsParamValue(GAMSParameterRecord param, double val, int places) {
		double dOut = Math.round(val * Math.pow(10,places)) / Math.pow(10,places);
		param.setValue(dOut);
		return dOut;
	}
		
	private GamsDemandOutput handleResults(GAMSDatabase outDB) {
		int modelStatus = (int) outDB.getParameter("ms").findRecord().getValue();
		String status = GAMSGlobals.ModelStat.lookup(modelStatus).toString();
		LogWriter.println(String.format("\nDemand %s: Modelstatus (%d) %s, Solvestatus %s", inputData.getCountry(),
				modelStatus, status, GAMSGlobals.SolveStat.lookup((int) outDB.getParameter("ss").findRecord().getValue()) ));
		
		GAMSVariable varU = outDB.getVariable("u");
		double utility = varU.getFirstRecord().getLevel();
		LogWriter.println(String.format("u = %.4f", utility)); 
		
		GAMSVariable varSubs = outDB.getVariable("SubsistenceC");
		GAMSVariable varDisc = outDB.getVariable("DiscretionaryC");

		HashSet<GamsCommodityDemand> demandMap = new HashSet<GamsCommodityDemand>();
		Collection<CommodityType> unhandledComms = new ArrayList<CommodityType>(CommodityType.getAllFoodItems());
		
		for (GAMSVariableRecord rec : varSubs) {
			String commodityName = rec.getKeys()[0];
			CommodityType commodity = CommodityType.getForGamsName(commodityName, true);
			
			if (commodity == null)
				LogWriter.println(String.format("Can't find commodity from Gams demand, so ignoring: " + commodityName)); 
			else {
				double subsGams = rec.getLevel();
				double discGams = modelStatus==0 ? 0 : varDisc.findRecord(commodityName).getLevel();  //modelStatus==0 is optimiser not run, i.e. too low GDP
				GamsCommodityDemand demand = new GamsCommodityDemand(commodity, subsGams, discGams,inputData.getKcalPerT(commodity));
				LogWriter.println(demand.toString());
				demandMap.add(demand);
				unhandledComms.remove(commodity);
			}
		}
		
		for (CommodityType commodity : unhandledComms) {
			LogWriter.println(commodity + " not found in Gams output so adding with zero demand"); 
			demandMap.add(new GamsCommodityDemand(commodity, 0, 0, 0));
		}
		
		GAMSParameter parConsumpFact = outDB.getParameter("desiredConsumpFactor");
		double desiredConsumpFactor = parConsumpFact.getFirstRecord().getValue();
		LogWriter.println(String.format("desiredConsumpFactor=%.4f", desiredConsumpFactor));
		
		return new GamsDemandOutput(status, demandMap, utility, desiredConsumpFactor);
	}
}
