package ac.ed.lurg.demand;

import java.io.BufferedWriter;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.country.CompositeCountryManager;
import ac.ed.lurg.country.SingleCountry;
import ac.ed.lurg.country.gams.GamsCommodityDemand;
import ac.ed.lurg.country.gams.GamsDemandInput;
import ac.ed.lurg.country.gams.GamsDemandOptimiser;
import ac.ed.lurg.country.gams.GamsDemandOutput;
import ac.ed.lurg.types.CommodityType;
import ac.ed.lurg.utils.FileWriterHelper;
import ac.ed.lurg.utils.LogWriter;

public class ElasticDemandManager extends AbstractSSPDemandManager {

	private Map<SingleCountry, Map<CommodityType, Double>> baseCpcCache = new HashMap<SingleCountry, Map<CommodityType, Double>>();
	private Map<SingleCountry, GamsDemandOutput> previousGamsDemands = new HashMap<SingleCountry, GamsDemandOutput>();
	private BufferedWriter outputFile;
	private PriceMarkUps priceMarkups;

	public ElasticDemandManager(String ssp_scenario, BaseConsumpManager baseConsumpManager, CalorieManager calorieManager, CompositeCountryManager compositeCountryManager) {
		super(ssp_scenario, baseConsumpManager, calorieManager, compositeCountryManager);
		priceMarkups = new PriceMarkUps();

		try {
			outputFile = FileWriterHelper.getFileWriter(true, true, ModelConfig.DEMAND_OPTIMISATION_OUTPUT_FILE, getDamsDemandOutputsHeader());
		} catch (IOException e) {
			LogWriter.print(e);
		}
	}

	@Override
	Map<CommodityType, Double> getFoodDemandMap(SingleCountry c, int year, double gdpPc, double population, Map<CommodityType, Double> producerPrices, double usaGdpPc) {
		Map<CommodityType, Double> consumerPrices = new HashMap<CommodityType, Double>();

		Map<CommodityType, Double> kcalPerT = calorieManager.get(c);
		
		double nonfoodPrice = 45.376+53.036*gdpPc/usaGdpPc;
		consumerPrices.put(CommodityType.NONFOOD, nonfoodPrice);
		
		// convert from producer to consumer prices
		for (CommodityType commodity : CommodityType.getAllFoodItems()) {
			double producerPricePlum = producerPrices.get(commodity);
			double producerPrice = producerPricePlum * (1000/kcalPerT.get(commodity) * 2000 *365);  //price per person per year commodity x per year assuming 2000 kcal a day

			double consumerPrice = priceMarkups.markedupPrice(c, commodity, producerPrice);
						
			consumerPrices.put(commodity, consumerPrice);
			LogWriter.println("Consumer price for " + commodity.getGamsName() + " is " + consumerPrice);

		}
		
		GamsDemandInput inputData = new GamsDemandInput(c, year, gdpPc, consumerPrices,kcalPerT, usaGdpPc, previousGamsDemands.get(c));

		// Do the projection
		GamsDemandOutput gamsOutput = null;
		for (int i = 1; i < 50; ++i) {
			gamsOutput = new GamsDemandOptimiser(inputData).getDemandPc();
			if (gamsOutput.getStatus().equals("SOLVED") || gamsOutput.getStatus().equals("UNDEFINED_STAT"))
				break;
			else {
				LogWriter.printlnError(i + ": Problem solving " + c + ", " + year + " got " + gamsOutput.getStatus());
				gamsOutput.resetForNewAttempt();
				inputData = new GamsDemandInput(c, year, gdpPc, consumerPrices, kcalPerT,usaGdpPc, gamsOutput);
			}
		}
		
		previousGamsDemands.put(c, gamsOutput);
		Map<CommodityType, Double> foodPc = gamsOutput.getPlumDemands();

		Map<CommodityType, Double> foodDemands = new HashMap<CommodityType, Double>();
		SspData baseSspData = sspManager.get(ssp_scenario, ModelConfig.BASE_YEAR, c);
		double baseGdpPc = baseSspData.getGdpPc();

		Map<CommodityType, Double> baseExpCpcMap = baseCpcCache.get(c);
		if (baseExpCpcMap == null) {  // we should only do this in the first year
			baseCpcCache.put(c, foodPc);
			baseExpCpcMap = foodPc;
		}

		for (Map.Entry<CommodityType, Double> entry : foodPc.entrySet()) {
			CommodityType commodity = entry.getKey();
			double minTonIntake = (commodity != CommodityType.NONFOOD) ? commodity.getMinFAOCalorieIntake()/kcalPerT.get(commodity) : 0.0;
			double newExpCpc = entry.getValue();
			double baseExpCpc = baseExpCpcMap.get(commodity);
			double baseCpc = baseConsumpManager.get(c, commodity);

			double cpc = rebaseConsumption(baseGdpPc, baseCpc, baseExpCpc, gdpPc, newExpCpc);
			double cpcAdj = (cpc >= minTonIntake) ? cpc : minTonIntake;
			double d = cpcAdj * population;
			foodDemands.put(commodity, d);
		}

		writeGamsDemandOutputs(inputData, gamsOutput, foodDemands, population);

		return foodDemands;
	}

	private String getDamsDemandOutputsHeader() {
		return "country,year,gdpPc,population,status,utility,hungerFactor,commodity,price,subsistence,discretionary,plumNotRebase,plumRebased,rebasedKcal";
	}

	private void writeGamsDemandOutputs(GamsDemandInput inputData, GamsDemandOutput gamsOutput, Map<CommodityType, Double> foodDemands, double population) {
		try {
			Map<CommodityType, Double> plumDemands = gamsOutput.getPlumDemands();
			String initialData = String.format("%s,%d,%.2f,%.6f,%s,%.6f,%.6f", 
					inputData.getCountry(), inputData.getYear(),inputData.getGdpPc(), population, 
					gamsOutput.getStatus(), gamsOutput.getUtility(), gamsOutput.getHungerFactor());
			
			for (CommodityType commodity : CommodityType.values()) {
				StringBuffer sbData = new StringBuffer(initialData);
				GamsCommodityDemand gamsDemand = gamsOutput.getGamsDemands(commodity);
				sbData.append(String.format(",%s,%.4f", commodity.getGamsName(), inputData.getPrices().get(commodity)));
				sbData.append(String.format(",%.4f,%.4f", gamsDemand.getSubsistence(), gamsDemand.getDiscretionary()));
				sbData.append(String.format(",%.4f,%.4f,%.4f", plumDemands.get(commodity), foodDemands.get(commodity)/population, 
						foodDemands.get(commodity)/population/365*inputData.getKcalPerT(commodity)));
				outputFile.write(sbData.toString());
				outputFile.newLine();
			}
		}
		catch (IOException e) {
			LogWriter.print(e);
		}
	}
}